package exemple_hmm;


/*
 * By Alexandre Savard
 * McGill University
 *
 *run with:
 *java -jar exemple_hmm.jar
 *
 * Based on a code by Jean-Marc Francois
 */

import java.util.*;
import java.util.ArrayList;

import be.ac.ulg.montefiore.run.jahmm.*;
import be.ac.ulg.montefiore.run.jahmm.learn.BaumWelchLearner;
import be.ac.ulg.montefiore.run.jahmm.toolbox.KullbackLeiblerDistanceCalculator;
import be.ac.ulg.montefiore.run.jahmm.toolbox.MarkovGenerator;


/**
 * This Class demonstrate how to build a new HMM with known parameters
 * implementing the Urn Ball model. It implements also a BaumWelchLearner 
 * instantiated with generated sequences from the first built HMM. A second
 * HMM is learnt from the first one using the iteration process and sequences 
 * generated by the first HMM. Distance is calculated between both HMM.  
 *  
 */ 

public class Main {
	
	/* Possible observation possible for this HMM */
        public enum ObservationType {
		BLUE, GREEN, YELLOW, RED, ORANGE;
		
		public ObservationDiscrete<ObservationType> observation() {
			return new ObservationDiscrete<ObservationType>(this);
		}
	};
        //public class Packet extends ArrayList{};
       
	//Main need to be static
	static public void main(String[] argv) 
	throws java.io.IOException {
		
		/* Build a HMM and generate observation 
                   sequences using this HMM */
                //
		
		Hmm<ObservationDiscrete<ObservationType>,?> hmm = buildHmm();
                
                /* Print out some information about this HMM  */
                // State transition probability Matrix
                System.out.println();
                System.out.println("State transition probability Matrix : ");
                for (int i=0; i<hmm.nbStates(); i++) {
                    for (int j=0; j<hmm.nbStates(); j++) {
                        if (j==hmm.nbStates()-1)
                            System.out.println("(" + hmm.getAij(i,j) + ")");
                        else
                            System.out.print("(" + hmm.getAij(i,j) + ")");
                    
                    }
                }
                System.out.println();
                // Observation Probability Distribution
                for (int i=0; i<hmm.nbStates(); i++) {
                    System.out.println();
                    System.out.println("Observation Probability Distribution " +
                            "Function for state "+ i);
                    System.out.println(hmm.getOpdf(i));
                }
                //Observation Probability Distribution
                System.out.println();
                for (int i=0; i<hmm.nbStates(); i++) {
                    System.out.println("Initial state probability(" 
                            + i + ") = " + hmm.getPi(i));
                }
                System.out.println();
		
		
                /* generate sequences from previous hmm */
                
                List<List<ObservationDiscrete<ObservationType>>> sequences;
		sequences = generateSequences(hmm);
		System.out.println();
                
                System.out.println();
                //System.out.println("Genereted Sequence :" + sequences);
                
		
		/* Baum-Welch learning */
                /* BaulmWelchLearner is generic, as well as the whole library 
                 * and needs to be instantiated 
                 * specifying the datatypes to be handled.
                 * The Constructor needs as a second element  
                 * a List.(here "sequence")
                 *
                 * List is an interface That handles collection. Needs to be
                 * extended to be used. 
                 *
                 * In Java, generics type correctness are checked at compile 
                 * time. An important drawback is that the generic type 
                 * information is not known at runtime. So, 
                 * List<ObservatioDiscrete> and a List<ObservationVector> are 
                 * refer seen as the same class, List.
                 *
                 * It is not possible to create an array of generic type 
                 * since it is impossible to evaluate its type. Then, once 
                 * created, an observation sequence cannot be modified. 
                 *
                 * In an environment such as Weka, the Lybrary needs to be 
                 * modified. "DynamicHmm"
                 * is a try to make the BaulmWelchLearner dynamic replacing
                 * all List<List<Observation>> by ArrayList. ArrayList is 
                 * a direct extension of the List interface 
                 */ 
                  
		
		BaumWelchLearner<ObservationDiscrete<ObservationType>, ?> bwl =
		   new BaumWelchLearner<ObservationDiscrete<ObservationType>,
		   OpdfDiscrete<ObservationType>>(
		      new OpdfDiscreteFactory<ObservationType>(
                          ObservationType.class),sequences);
		
		
                //Initialize the HMM tha twill be used in the itterative process
                Hmm<ObservationDiscrete<ObservationType>,?> learntHmm = 
                        buildInitHmm();
		
		// This object measures the distance between two HMMs
		KullbackLeiblerDistanceCalculator klc = 
			new KullbackLeiblerDistanceCalculator();
		
		// Incrementally improve the solution
		for (int i = 0; i < 5; i++) {
			System.out.println("Distance at iteration " + i + ": " +
					klc.distance(learntHmm, hmm));
			learntHmm = bwl.iterate(learntHmm);
		}
		
		
		/* Computing the probability of a sequence */
		
		ObservationDiscrete<ObservationType> obsBlue = 
                                       ObservationType.BLUE.observation();
		ObservationDiscrete<ObservationType> obsGreen = 
                                       ObservationType.GREEN.observation();
                ObservationDiscrete<ObservationType> obsYellow = 
                                       ObservationType.YELLOW.observation();
		ObservationDiscrete<ObservationType> obsRed = 
                                       ObservationType.RED.observation();
                ObservationDiscrete<ObservationType> obsOrange = 
                                       ObservationType.ORANGE.observation();
		
		List<ObservationDiscrete<ObservationType>> testSequence; 
		testSequence = Arrays.asList(obsBlue, obsGreen, obsRed,
                                                obsYellow, obsOrange);
                
                System.out.println();
                System.out.println("Test Sequence :" + testSequence);
		
                System.out.println();
		System.out.println("Sequence probability from initial HMM: " +
				hmm.probability(testSequence));
                
                System.out.println();
		System.out.println("Sequence probability from learnt HMM: " +
				learntHmm.probability(testSequence));
                
                /* Print out some information about this HMM  */
                
                System.out.println();
                System.out.println("State transition probability " +
                                                        "distribution : ");
                for (int i=0; i<learntHmm.nbStates(); i++) {
                    for (int j=0; j<learntHmm.nbStates(); j++) {
                        if (j==learntHmm.nbStates()-1)
                            System.out.println("(" + 
                                                  learntHmm.getAij(i,j) + ")");
                        else
                            System.out.print("(" + learntHmm.getAij(i,j) + ")");
                    
                    }
                }
                System.out.println();
                
                for (int i=0; i<learntHmm.nbStates(); i++) {
                    System.out.println();
                    System.out.println("Observation Probability Distribution " +
                            "Function for state "+ i);
                    System.out.println(learntHmm.getOpdf(i));
                }
                
                System.out.println();
                for (int i=0; i<learntHmm.nbStates(); i++) {
                    System.out.println("Pi(" + i + ") = " + learntHmm.getPi(i));
                }
                System.out.println();
		
	}
	
	
	/* The HMM this example is based on */
	
	static Hmm<ObservationDiscrete<ObservationType>,?> buildHmm() {
		
		Hmm<ObservationDiscrete<ObservationType>, 
                        OpdfDiscrete<ObservationType>> hmm = 
			   new Hmm<ObservationDiscrete<ObservationType>,
			        OpdfDiscrete<ObservationType>>(5,
			           new OpdfDiscreteFactory<ObservationType>(
                                        ObservationType.class));
		
		hmm.setPi(0, 0.10);
                hmm.setPi(1, 0.05);
                hmm.setPi(2, 0.20);
                hmm.setPi(3, 0.30);
                hmm.setPi(4, 0.35);
                

                hmm.setOpdf(0, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(1, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.02, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(2, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(3, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(4, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));

                hmm.setAij(0, 1, 0.25);
                hmm.setAij(0, 2, 0.25);
                hmm.setAij(0, 3, 0.25);
                hmm.setAij(0, 4, 0.25);
                
                hmm.setAij(1, 0, 0.25);
                hmm.setAij(1, 2, 0.25);
                hmm.setAij(1, 3, 0.25);
                hmm.setAij(1, 4, 0.25);
                
                hmm.setAij(2, 0, 0.25);
                hmm.setAij(2, 1, 0.25);
                hmm.setAij(2, 3, 0.25);
                hmm.setAij(2, 4, 0.25);
                
                hmm.setAij(3, 0, 0.25);
                hmm.setAij(3, 1, 0.25);
                hmm.setAij(3, 2, 0.25);
                hmm.setAij(3, 4, 0.25);
                
                hmm.setAij(4, 0, 0.25);
                hmm.setAij(4, 1, 0.25);
                hmm.setAij(4, 2, 0.25);
                hmm.setAij(4, 3, 0.25);
		
		return hmm;
	}
	
	
	/* Initial guess for the Baum-Welch algorithm */
	
	/* The HMM this example is based on */
	
	static Hmm<ObservationDiscrete<ObservationType>,?> buildInitHmm() {
		
		Hmm<ObservationDiscrete<ObservationType>, 
                     OpdfDiscrete<ObservationType>> hmm = 
			new Hmm<ObservationDiscrete<ObservationType>,
			   OpdfDiscrete<ObservationType>>(5,
                                new OpdfDiscreteFactory<ObservationType>
                                            (ObservationType.class));
		
                hmm.setPi(0, 0.20);
                hmm.setPi(1, 0.20);
                hmm.setPi(2, 0.20);
                hmm.setPi(3, 0.20);
                hmm.setPi(4, 0.20);
                

                hmm.setOpdf(0, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(1, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.02, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(2, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(3, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));
                hmm.setOpdf(4, new OpdfDiscrete<ObservationType>(
                        ObservationType.class,new double[] { 
                                        0.2, 0.2, 0.2, 0.2, 0.2 }));

                hmm.setAij(0, 1, 0.25);
                hmm.setAij(0, 2, 0.25);
                hmm.setAij(0, 3, 0.25);
                hmm.setAij(0, 4, 0.25);
                
                hmm.setAij(1, 0, 0.25);
                hmm.setAij(1, 2, 0.25);
                hmm.setAij(1, 3, 0.25);
                hmm.setAij(1, 4, 0.25);
                
                hmm.setAij(2, 0, 0.25);
                hmm.setAij(2, 1, 0.25);
                hmm.setAij(2, 3, 0.25);
                hmm.setAij(2, 4, 0.25);
                
                hmm.setAij(3, 0, 0.25);
                hmm.setAij(3, 1, 0.25);
                hmm.setAij(3, 2, 0.25);
                hmm.setAij(3, 4, 0.25);
                
                hmm.setAij(4, 0, 0.25);
                hmm.setAij(4, 1, 0.25);
                hmm.setAij(4, 2, 0.25);
                hmm.setAij(4, 3, 0.25);
		
		return hmm;
	}
	
	
	/* Generate several observation sequences using a HMM */
	
	static <O extends Observation> List<List<O>> 
	generateSequences(Hmm<O,?> hmm) {
	
		MarkovGenerator<O> mg = new MarkovGenerator<O>(hmm);
		
		List<List<O>> sequences = new ArrayList<List<O>>();
		for (int i = 0; i < 200; i++)
			sequences.add(mg.observationSequence(100));

		return sequences;
	}

}
